from hestia_earth.utils.request import join_args, request_url


def test_join_args():
    assert join_args(['arg1', '', 'arg3']) == 'arg1&arg3'


def test_request_url():
    assert request_url('base', id='id', empty='') == 'base?id=id'



from datetime import datetime
import numpy as np
from hestia_earth.schema import NodeType

from hestia_earth.utils.tools import (
    non_empty_value,
    non_empty_list,
    is_term,
    current_time_ms,
    safe_parse_float,
    safe_parse_date,
    list_sum,
    list_average,
    flatten,
    get_dict_key,
    to_precision,
    is_number,
    is_boolean,
    omit,
    pick,
    unique_values
)


def test_non_empty_value():
    assert not non_empty_value('')
    assert not non_empty_value([])
    assert not non_empty_value({})
    assert non_empty_value('test') is True
    assert non_empty_value(False) is True
    assert non_empty_value(1) is True
    assert non_empty_value(np.float64(10)) is True


def test_non_empty_list():
    assert non_empty_list(['', 1, [], False]) == [1, False]


def test_is_term():
    assert not is_term({'@type': NodeType.CYCLE.value})
    assert is_term({'type': NodeType.TERM.value}) is True


def test_current_time_ms():
    assert current_time_ms() > 1000


def test_safe_parse_float():
    assert safe_parse_float('123.456') == 123.456
    assert safe_parse_float('abcd', 10) == 10
    assert safe_parse_float(np.nan, 1) == 1


def test_safe_parse_date():
    assert safe_parse_date('2020-01-01') == datetime(2020, 1, 1)
    assert safe_parse_date('abcd', datetime(2020, 1, 1)) == datetime(2020, 1, 1)


def test_list_sum():
    assert list_sum([0, 1]) == 1
    assert list_sum([]) == 0
    assert list_sum(None) == 0
    assert list_sum('value') == 0

    assert list_sum([], None) is None
    assert list_sum([None], None) is None

    assert list_sum(['random'], None) is None


def test_list_average():
    assert list_average([0, 1]) == 0.5
    assert list_average([]) == 0
    assert list_average(None) == 0
    assert list_average('value') == 0

    assert list_average([], None) is None
    assert list_average([None], None) is None

    assert list_average(['random'], None) is None


def test_flatten():
    assert flatten([[0, 1], 2, [3, 4]]) == [0, 1, 2, 3, 4]


def test_get_dict_key():
    value = {
        'a': 'test',
        'b': {
            'id': 'id'
        },
        'c': {
            'd': [
                {
                    'id': 1
                },
                {
                    'id': 2
                }
            ]
        }
    }
    assert get_dict_key(value, 'a') == 'test'
    assert get_dict_key(value, 'b.id') == 'id'
    assert get_dict_key(value, 'c.d.id') == [1, 2]


def test_to_precision():
    assert to_precision(0.45249) == 0.452
    assert to_precision(1.45213) == 1.45
    assert to_precision(144.5213) == 145

    assert to_precision(0.45249, 1) == 0.5
    assert to_precision(1.45213, 1) == 1
    assert to_precision(145, 1) == 100

    assert to_precision(0.000152, 3) == 0.000152
    assert to_precision(9089080.000000001111, 3) == 9090000


def test_is_number():
    assert is_number(10) is True
    assert is_number(10.01) is True
    assert not is_number(True)
    assert not is_number(False)
    assert not is_number('test')
    assert not is_number('0')


def test_is_boolean():
    assert not is_boolean(10)
    assert not is_boolean(10.01)
    assert is_boolean(True) is True
    assert is_boolean(False) is True
    assert not is_boolean('test')
    assert not is_boolean('0')


def test_omit():
    assert omit({'a': 1, 'b': 2}, ['a']) == {'b': 2}


def test_pick():
    assert pick({'a': 1, 'b': 2}, ['a']) == {'a': 1}


def test_unique_values():
    values = [{'@id': 1, 'name': 'a'}, {'@id': 2, 'name': 'a'}]
    assert unique_values(values) == values

    values = [{'@id': 1, 'name': 'a'}, {'@id': 1, 'name': 'b'}]
    assert unique_values(values) == [{'@id': 1, 'name': 'b'}]



import os

from .tools import non_empty_value


def api_url() -> str: return os.getenv('API_URL', 'https://api.hestia.earth')


def api_access_token() -> str: return os.getenv('API_ACCESS_TOKEN')


def web_url() -> str: return os.getenv('WEB_URL', 'https://www.hestia.earth')


def join_args(values) -> str: return '&'.join(list(filter(non_empty_value, values))).strip()


def request_url(base_url: str, **kwargs) -> str:
    args = list(map(lambda key: '='.join([key, str(kwargs.get(key))]) if kwargs.get(key) else None, kwargs.keys()))
    return '?'.join(list(filter(non_empty_value, [base_url, join_args(args)]))).strip()



from os.path import join
import json
import numpy as np

from .tools import current_time_ms, non_empty_list, flatten
from .api import find_related
from .storage._s3_client import (
    _load_from_bucket, _upload_to_bucket, _last_modified, _read_metadata, _update_metadata, _exists_in_bucket
)
from .storage._sns_client import _get_sns_client

PROGRESS_EXT = '.progress'
CALC_FOLDER = 'recalculated'
METADATA_STAGE_KEY = 'stage'
METADATA_PROGRESS_KEY = 'calculating'


# fix error "Object of type int64 is not JSON serializable"
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


def to_string(data: dict, indent: int = None): return json.dumps(data, indent=indent, ensure_ascii=False, cls=NpEncoder)


def to_bytes(data: dict): return to_string(data).encode('utf8')


def upload_json(bucket_name: str, file_key: str, body: dict):
    return _upload_to_bucket(
        bucket=bucket_name,
        key=file_key,
        body=to_bytes(body),
        content_type='application/json'
    )


def _to_file_progress(filepath: str):
    return filepath.replace('.csv', PROGRESS_EXT).replace('.json', PROGRESS_EXT).replace('.hestia', PROGRESS_EXT)


def handle_result(bucket_name: str, file_key: str, step: str, start: int, content: dict):
    filepath = _to_file_progress(file_key)

    # try to read existing progress to update the time per step
    try:
        data = json.loads(_load_from_bucket(bucket_name, filepath))
    except Exception:
        data = {}

    return upload_json(bucket_name, filepath, {
        **data,
        'step': step,
        'time': {
            **(data.get('time', {}) if isinstance(data.get('time', {}), dict) else {}),
            step: current_time_ms() - start
        },
        **content
    })


def handle_error(
    bucket_name: str, file_key: str, step: str, start: int,
    err: str = '', stack: str = '', errors=[], warnings=[],
    extras: dict = {}
):
    return handle_result(bucket_name, file_key, step, start, extras | {
        'success': False,
        'error': {
            'message': err,
            'stack': stack,
            'errors': errors,
            'warnings': warnings
        },
        'warning': {
            'warnings': warnings
        }
    })


def handle_success(bucket_name: str, file_key: str, step: str, start: int, extras: dict = {}):
    return handle_result(bucket_name, file_key, step, start, extras | {'success': True})


def publish_result(topic_arn: str, bucket_name: str, file_key: str, filepath: str, step: str, success: bool):
    return _get_sns_client().publish(
        TopicArn=topic_arn,
        Message=to_string({
            'bucket': bucket_name,
            'key': file_key,
            'filepath': filepath
        }),
        MessageAttributes={
            'functionName': {
                'DataType': 'String',
                'StringValue': step + ('Done' if success else 'Error')
            }
        }
    )


def _parse_event_s3(event: dict):
    return {'bucket': event['s3']['bucket']['name'], 'key': event['s3']['object']['key']}


def _parse_event_SNS(event: dict):
    event = event.get('Sns', {})
    data = json.loads(event.get('Message', '{}'))
    attributes: dict = event.get('MessageAttributes', {})
    data['attributes'] = {key: value.get('Value') for key, value in attributes.items()}
    return data


def _parse_event_SQS(event: dict):
    condition = event.get('requestContext', {}).get('condition')
    return _get_data_from_event(event.get('requestPayload', {})) if condition != 'RetriesExhausted' else None


def _get_data_from_event(event):  # noqa: C901
    if isinstance(event, dict):
        if 's3' in event:
            return _parse_event_s3(event)
        # invoked when running asynchronously
        if 'Sns' in event:
            return _parse_event_SNS(event)
        # invoked through http event
        if 'body' in event:
            return _get_data_from_event(json.loads(event.get('body', '{}')))
        # invoked through s3 put object
        if 'Records' in event:
            return flatten(map(_get_data_from_event, event.get('Records', [])))
        # invoked when calculation timedout or failed
        if 'requestPayload' in event:
            return _parse_event_SQS(event)
        return event
    if isinstance(event, str):
        return _get_data_from_event(json.loads(event))


def parse_event(event: dict):
    data = _get_data_from_event(event)
    return non_empty_list(flatten(data) if isinstance(data, list) else [data])


def _node_type(node: dict): return node.get('@type', node.get('type'))


def _node_id(node: dict): return node.get('@id', node.get('id'))


def _node_path(node: dict, folder: str = ''): return join(folder, _node_type(node), f"{_node_id(node)}.jsonld")


def _load_node(bucket: str, file_key: str): return json.loads(_load_from_bucket(bucket, file_key))


def _cache_path(node: dict): return join(_node_type(node), f"{_node_id(node)}.cache")


def _has_cache(bucket: str, node: dict): return _exists_in_bucket(bucket, _cache_path(node))


def is_calculating(bucket: str, node: dict, folder: str = ''):
    return _read_metadata(bucket, _node_path(node, folder)).get(METADATA_PROGRESS_KEY, 'false') == 'true'


def set_calculating(bucket: str, node: dict, in_progress: bool, folder: str = ''):
    return _update_metadata(bucket, _node_path(node, folder), {METADATA_PROGRESS_KEY: str(in_progress).lower()})


def get_stage(bucket: str, node: dict, folder: str = CALC_FOLDER):
    stage = _read_metadata(bucket, _node_path(node, folder=CALC_FOLDER)).get(METADATA_STAGE_KEY)
    return int(stage) if stage else stage


def load_cache(bucket: str, node: dict):
    """
    Return the cache data for the node.

    Parameters
    ----------
    bucket : str
        The bucket where the cache is stored.
    node : dict
        The Node which is connected to other nodes (source).

    Returns
    -------
    dict
        The cached data.
    """
    cache_path = join(node['@type'], f"{node['@id']}.cache")
    try:
        return json.loads(_load_from_bucket(bucket, cache_path))
    except Exception:
        print('No cache found for', cache_path)
        return {}


def _filter_by_type(nodes: list, type: str): return [n for n in nodes if n.get('@type', n.get('type')) == type]


def _find_related_nodes(from_type: str, from_id: str, related_type: str, related_key: str):
    should_find_related = related_key == 'related'
    print('Find related nodes from API', from_type, from_id, related_key, related_type)
    return find_related(from_type, from_id, related_type, limit=10000) if should_find_related else []


def _get_cached_nodes(cache: dict, related_key: str, from_type: str, from_id: str, to_type: str):
    # if key is in cache, use nodes in cache, otherwise use API
    if related_key in cache:
        nodes = _filter_by_type(cache.get(related_key, []), to_type)
        print('Using cached data to', related_key, to_type, nodes)
        return list(map(lambda node: {'@type': to_type, '@id': node.get('@id', node.get('id'))}, nodes))
    else:
        return _find_related_nodes(from_type, from_id, to_type, related_key)


def get_related_nodes(node: dict, related_key: str, related_type: str, cache: dict):
    """
    Given a node, return all related nodes.

    Parameters
    ----------
    node : dict
        The Node which is connected to other nodes (source).
    related_key : str
        Either `nested` or `related`.
    related_type : str
        Related node `@type`.
    cache : dict
        Cache data of the source Node. Can contain `related` and `nested` nodes.
        When no provided, only `related` nodes can be found.

    Returns
    -------
    List[dict]
        The related nodes.
    """
    from_type = node.get('@type', node.get('type'))
    from_id = node.get('@id', node.get('id'))

    related_nodes = _get_cached_nodes(cache or {}, related_key, from_type, from_id, related_type) or []

    return list({f"{node['@type']}/{node['@id']}": node for node in related_nodes}.values())


def get_related_nodes_data(bucket_name: str, node: dict, related_key: str, related_type: str, cache: dict):
    """
    Given a node, return all related nodes with extra data.

    Parameters
    ----------
    bucket_name : str
        The name of the bucket where the nodes are stored.
    node : dict
        The Node which is connected to other nodes (source).
    related_key : str
        Either `nested` or `related`.
    related_type : str
        Related node `@type`.
    cache : dict
        Cache data of the source Node. Can contain `related` and `nested` nodes.
        When no provided, only `related` nodes can be found.

    Returns
    -------
    List[dict]
        The related nodes with extra data: `indexed_at`, `recalculated_at` and `recalculated_stage`.
    """
    related_nodes = get_related_nodes(node=node, related_key=related_key, related_type=related_type, cache=cache)

    return [
        node | {
            'indexed_at': _last_modified(bucket=bucket_name, key=_node_path(node)),
            'recalculated_at': _last_modified(bucket=bucket_name, key=_node_path(node, folder=CALC_FOLDER)),
            'recalculated_stage': get_stage(bucket_name, node),
            'is_calculating': is_calculating(bucket_name, node)
        }
        for node in related_nodes
    ]



from numpy import array
from pytest import mark
from hestia_earth.schema import MeasurementStatsDefinition

from hestia_earth.utils.descriptive_stats import calc_descriptive_stats

EXPECTED_FLATTENED = {
    "value": [5],
    "sd": [2.581989],
    "min": [1],
    "max": [9],
    "statsDefinition": "simulated",
    "observations": [9]
}

EXPECTED_COLUMNWISE = {
    "value": [4, 5, 6],
    "sd": [2.44949, 2.44949, 2.44949],
    "min": [1, 2, 3],
    "max": [7, 8, 9],
    "statsDefinition": "simulated",
    "observations": [3, 3, 3]
}

EXPECTED_ROWWISE = {
    "value": [2, 5, 8],
    "sd": [0.816497, 0.816497, 0.816497],
    "min": [1, 4, 7],
    "max": [3, 6, 9],
    "statsDefinition": "simulated",
    "observations": [3, 3, 3]
}


@mark.parametrize(
    "axis, expected",
    [(None, EXPECTED_FLATTENED), (0, EXPECTED_COLUMNWISE), (1, EXPECTED_ROWWISE)],
    ids=["flattened", "columnwise", "rowwise"]
)
@mark.parametrize("stats_definition", [MeasurementStatsDefinition.SIMULATED, "simulated"], ids=["Enum", "str"])
def test_calc_descriptive_stats(stats_definition, axis, expected):
    ARR = array([
        [1, 2, 3],
        [4, 5, 6],
        [7, 8, 9]
    ])

    result = calc_descriptive_stats(ARR, stats_definition, axis=axis)
    assert result == expected



import json
from functools import lru_cache
from typing import Union
from hestia_earth.schema import TermTermType

from .storage import _load_from_storage
from .api import download_hestia


@lru_cache()
def _load_term_file(term_type: str):
    try:
        filepath = f"glossary/{term_type}.json"
        nodes = json.loads(_load_from_storage(filepath, glossary=True))
        return {node.get('@id'): node for node in nodes}
    except Exception:
        return {}


def download_term(term: Union[str, dict], termType: Union[str, TermTermType] = None):
    """
    Download a Term, using the glossary file if available, or default to the standard download.
    """
    term_id = term.get('@id', term.get('id')) if isinstance(term, dict) else term
    term_type = (
        termType if isinstance(termType, str) else termType.value
    ) if termType else (
        term.get('termType') if isinstance(term, dict) else None
    )
    cached_nodes = _load_term_file(term_type) if term_type else {}
    return cached_nodes.get(term_id) or download_hestia(term_id)



# import os
# import json
# from pytest import mark
# from hestia_earth.schema import TermTermType

# from tests.utils import fixtures_path
# from hestia_earth.utils.blank_node import get_node_value, ArrayTreatment, get_blank_nodes_calculation_status

# fixtures_folder = os.path.join(fixtures_path, 'blank_node')
# calculation_status_folder = os.path.join(fixtures_folder, 'calculation_status')


# def test_get_node_value():
#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': [10]}
#     assert get_node_value(blank_node, 'value', default=None) == 10

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': [0]}
#     assert get_node_value(blank_node, 'value', default=None) == 0

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}}
#     assert get_node_value(blank_node, 'value', default=None) is None

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': [10, 20]}
#     assert get_node_value(blank_node, 'value', default=None) == 15

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': True}
#     assert get_node_value(blank_node, 'value', default=None) is True

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': 10}
#     assert get_node_value(blank_node, 'value', default=None) == 10

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': None}
#     assert get_node_value(blank_node, 'value', default=None) is None

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': None}
#     assert get_node_value(blank_node, 'value', default=0) == 0

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': [10, None, 20]}
#     assert get_node_value(blank_node, 'value', default=None) == 15

#     blank_node = {'term': {'termType': 'crop', '@id': 'wheatGrain'}, 'value': [10, None, 20]}
#     assert get_node_value(blank_node, 'value', array_treatment=ArrayTreatment.SUM) == 30


# @mark.parametrize(
#     'folder,list_key,termType',
#     [
#         ('cycle', 'emissions', TermTermType.EMISSION),
#     ]
# )
# def test_get_blank_nodes_calculation_status(folder: str, list_key: str, termType: TermTermType):
#     with open(f"{calculation_status_folder}/{folder}/node.jsonld", encoding='utf-8') as f:
#         node = json.load(f)

#     with open(f"{calculation_status_folder}/{folder}/{list_key}-{termType.value}.json", encoding='utf-8') as f:
#         expected = json.load(f)

#     result = get_blank_nodes_calculation_status(node, list_key=list_key, termType=termType)
#     assert result == expected, folder



import pandas as pd

from .utils import fixtures_path
from hestia_earth.utils.table import format_for_upload


def test_format_for_upload():
    filepath = f"{fixtures_path}/table/format/sample.csv"
    expected = pd.read_csv(f"{fixtures_path}/table/format/output.csv", index_col=None)
    df = format_for_upload(filepath)
    assert df.to_string() == expected.to_string()



import copy
import json
import re
import numpy as np
from hestia_earth.schema import UNIQUENESS_FIELDS, Term, NODE_TYPES
from hestia_earth.schema.utils.sort import get_sort_key, SORT_CONFIG
from flatten_json import flatten as flatten_json

# __package__ = "hestia_earth.utils" # required to run interactively in vscode
from ..api import find_term_ids_by_names
from ._shared import EXCLUDE_FIELDS, EXCLUDE_PRIVATE_FIELDS, _with_csv_formatting, _filter_emissions_not_relevant


PANDAS_IMPORT_ERROR_MSG = "Run `pip install pandas>=1.2` to use this functionality"
try:
    import pandas as pd

    version = [int(x) for x in pd.__version__.split('+')[0].split(".")]
    if version[0] < 1 or (version[0] == 1 and version[1] < 2):
        raise ImportError(PANDAS_IMPORT_ERROR_MSG)
except ImportError:
    raise ImportError(PANDAS_IMPORT_ERROR_MSG)


# We only want to pivot array items containing blank nodes
# Assume these are all fields with uniqueness fields not of type Node
def _get_blank_node_uniqueness_fields():
    filtered_uniqueness_fields = copy.deepcopy(UNIQUENESS_FIELDS)
    for node_type, array_fields in UNIQUENESS_FIELDS.items():
        for array_field in array_fields.keys():
            if SORT_CONFIG[node_type][array_field]['type'] in NODE_TYPES:
                del filtered_uniqueness_fields[node_type][array_field]
            # include `impactAssessment.@id` since it is not part of original uniqueness
            if 'impactAssessment.id' in array_fields[array_field]:
                filtered_uniqueness_fields[node_type][array_field].append('impactAssessment.@id')
    return filtered_uniqueness_fields


BLANK_NODE_UNIQUENESS_FIELDS = _get_blank_node_uniqueness_fields()


# assuming column labels always camelCase
def _get_node_type_label(node_type):
    return node_type[0].lower() + node_type[1:]


def _get_node_type_from_label(node_type):
    return node_type[0].upper() + node_type[1:]


def _get_names(df):
    names = []
    for node_type, array_fields in BLANK_NODE_UNIQUENESS_FIELDS.items():
        nt_label = _get_node_type_label(node_type)
        for array_field, uniqueness_fields in array_fields.items():
            regex = rf"^{nt_label}\.{array_field}\.\d+"
            field_cols = df.filter(regex=regex).copy()
            deep_cols = [col for col in field_cols.columns if _get_deep_col(col)]
            if len(deep_cols):
                deep_df = field_cols[deep_cols].rename(columns=_get_deep_col)
                deep_df.columns = pd.MultiIndex.from_tuples(deep_df.columns)
                for _original_prefix, term in deep_df.T.groupby(level=0):
                    deep_names = _get_names(term.T.droplevel(0, axis=1))
                    names.extend(deep_names)
            for field in [f for f in uniqueness_fields if ".@id" in f]:
                name_field = field.replace(".@id", ".name")
                regex = rf"^{nt_label}\.{array_field}\.\d+\.{name_field}$"
                names_for_field = field_cols.filter(regex=regex).to_numpy().flatten()
                names.extend(names_for_field)
    valid_names = [name for name in set(names) if pd.notna(name)]
    return valid_names


def _get_name_id_dict(df):
    names = _get_names(df)
    return find_term_ids_by_names(sorted(names))


def _sort_keys(index):
    return [get_sort_key(col) for col in index]


def _sort_inplace(df):
    df.sort_index(axis=1, inplace=True, key=_sort_keys)


def _get_term_field(col_header):
    return ".".join(col_header.split(".")[3:])


def _get_term_index(col_header):
    return f"term.{col_header.split('.')[2]}"


def _group_by_term(termT, name_id_dict, uniqueness_fields):
    term = termT.T
    term.columns = map(_get_term_field, term.columns)

    # fill in any missing ids prior to grouping by id
    replace_fields = {
        f: f.replace(".@id", ".name") for f in uniqueness_fields if ".@id" in f
    }
    for idField, nameField in replace_fields.items():
        term[idField] = term.apply(
            lambda row: row.get(
                idField,
                name_id_dict.get(row.get(nameField, "default_id_value"), np.nan),
            ),
            axis=1,
        )
        term.drop(nameField, axis=1, inplace=True, errors="ignore")
    return term.T


def _format_col_field(field):
    return field.split(".")[0]


def _pivot_by_term(row_tuple, columns, fields_to_include):
    row = {k: v for k, v in zip(columns, row_tuple)}
    value_col_fields = []
    series = {}
    for col, append_to_value_col in fields_to_include.items():
        if pd.notna(row[col]):
            if append_to_value_col:
                value_col_fields.append(col)
            else:
                series[f"{row['term.@id']}.{col}"] = row[col]
    value_col_fields.sort()
    value_col_prefix = f"{row['term.@id']}{'+' if len(value_col_fields) else ''}"
    value_col_suffix = "+".join(
        [f"{_format_col_field(col)}[{row[col]}]" for col in value_col_fields]
    )
    value_col = f"{value_col_prefix}{value_col_suffix}.value"
    series[value_col] = row.get("value", np.nan)
    return series


def _is_not_term_field(col):
    deepest_field = col.split(".")[-1]
    is_name = (
        deepest_field == "name"
    )  # keep .name fields that are not being pivoted (because not a uniqueArrayItem)
    return is_name or deepest_field not in Term().fields


def _pivot_by_term_id_group(terms, uniqueness_fields):
    fields_to_include = {}
    for col in terms.columns:
        if col in uniqueness_fields:
            single_value = (
                terms[col].eq(terms[col].iloc[0]).all() or terms[col].isna().all()
            )
            append_to_value_col = (
                # depthUpper and depthLower are exceptions which go into value col no matter what
                col == "depthUpper"
                or col == "depthLower"
                or not single_value
            )
            fields_to_include[col] = append_to_value_col
        elif col != "value" and _is_not_term_field(col):
            fields_to_include[col] = False
    del fields_to_include["term.@id"]
    # print('terms->', terms, '\n')
    pivoted_arr = [
        _pivot_by_term(row_tuple, terms.columns, fields_to_include)
        for row_tuple in zip(*[terms[col] for col in terms])
    ]
    pivoted = pd.DataFrame.from_records(pivoted_arr, index=terms.index)
    # pivoted = terms.apply(_pivot_by_term, axis=1, fields_to_include=fields_to_include)
    # print('pivoted->', pivoted, '\n')
    return pivoted


def _pivot_row(row, uniqueness_fields):
    # unstack to group sets of values for each id, then stack to restore row
    # unstacking looks like this:
    #          term.@id dates	            startDate	value
    # term.0	gwp100	01-02-22;02-02-22	01-01-22	10;12
    # term.1	gwp100	01-02-22;02-02-22	02-01-22	20;30
    # term.2	someId	bla                 bla         1.0

    pivoted = (
        row.unstack()
        .groupby("term.@id", group_keys=False)
        .apply(_pivot_by_term_id_group, uniqueness_fields=uniqueness_fields)
    )
    return pivoted.stack(dropna=False)


def _get_deep_col(col):
    parent_type_label, field, idx, *rest = col.split(".")
    deep_col = ".".join(rest)
    match = re.search(r"\.\d+\.", deep_col)
    if match:
        node_type = (
            SORT_CONFIG.get(_get_node_type_from_label(parent_type_label))
            .get(field)
            .get("type")
        )
        original_prefix = ".".join([parent_type_label, field, idx])
        raised_deep_col = f"{_get_node_type_label(node_type)}.{deep_col}"
        return (original_prefix, raised_deep_col)
    else:
        return None


def _restore_deep_col_prefix(original_prefix):
    def do_restore(deep_col):
        original_suffix = deep_col.split(".")[1:]
        return f'{original_prefix}.{".".join(original_suffix)}'

    return do_restore


def _do_pivot(df_in, name_id_dict):
    for node_type, array_fields in BLANK_NODE_UNIQUENESS_FIELDS.items():
        nt_label = _get_node_type_label(node_type)
        for field in array_fields:
            regex = rf"^{nt_label}\.{field}\.\d+"
            uniqueness_fields = array_fields[field]
            field_cols = df_in.filter(regex=regex).copy()
            if field_cols.empty:
                continue
            deep_cols = [col for col in field_cols.columns if _get_deep_col(col)]
            if len(deep_cols):
                deep_df = field_cols[deep_cols].rename(columns=_get_deep_col)
                field_cols.drop(deep_cols, axis=1, inplace=True)
                deep_df.columns = pd.MultiIndex.from_tuples(deep_df.columns)
                for original_prefix, term in deep_df.T.groupby(level=0):
                    deep_pivoted = _do_pivot(term.T.droplevel(0, axis=1), name_id_dict)
                    deep_pivoted.rename(
                        columns=_restore_deep_col_prefix(original_prefix), inplace=True
                    )
                    field_cols = field_cols.merge(
                        deep_pivoted, left_index=True, right_index=True, how="outer"
                    )

            field_cols.dropna(axis=0, how='all', inplace=True)

            with_grouped_cols = field_cols.T.groupby(
                _get_term_index, group_keys=True
            ).apply(
                _group_by_term,
                name_id_dict=name_id_dict,
                uniqueness_fields=uniqueness_fields
            ).T

            pivoted_terms = with_grouped_cols.apply(
                _pivot_row, axis=1, uniqueness_fields=uniqueness_fields
            )

            # merge any duplicated columns caused by shuffled term positions
            # this operation coincidentally sorts the columns alphabetically
            pivoted_terms = pivoted_terms.T.groupby(
                level=pivoted_terms.columns.nlevels - 1, group_keys=False
            ).apply(lambda term: term.bfill().iloc[0, :]).T

            pivoted_terms.columns = map(
                lambda col: f"{nt_label}.{field}.{col}", pivoted_terms.columns
            )
            df_in.drop(df_in.filter(regex=regex).columns, axis=1, inplace=True)
            df_in = df_in.merge(
                pivoted_terms, left_index=True, right_index=True, how="outer"
            )
    return df_in


def _format_and_pivot(df_in):
    df_in.replace("-", np.nan, inplace=True)
    df_in.dropna(how="all", axis=1, inplace=True)
    df_in.drop(columns="-", errors="ignore", inplace=True)
    name_id_dict = _get_name_id_dict(df_in)

    df_out = _do_pivot(df_in.copy(), name_id_dict)

    _sort_inplace(df_out)
    df_out = df_out.astype('object')
    df_out.fillna("-", inplace=True)
    return df_out


def nodes_to_df(nodes: list[dict]):
    nodes_flattened = [
        flatten_json(dict([(_get_node_type_label(node.get("@type", node.get('type'))), node)]), '.') for node in nodes
    ]

    return pd.json_normalize(nodes_flattened)


def pivot_nodes(nodes: list[dict]):
    """
    Pivot array of nodes in dict format (e.g under the 'nodes' key of a .hestia file)
    The nodes json should first be parsed using _with_csv_formatting for output as csv.
    """
    df_in = nodes_to_df(list(map(_filter_emissions_not_relevant, nodes)))

    regex = rf".*\.({'|'.join(EXCLUDE_FIELDS + EXCLUDE_PRIVATE_FIELDS)})$"
    df_in.drop(df_in.filter(regex=regex).columns, axis=1, inplace=True)
    df_out = _format_and_pivot(df_in)
    return df_out


def pivot_hestia_file(hestia_file: str):
    """
    Parse a .hestia file and pivot the nodes within.

    Parameters
    ----------
    hestia_file : str
        A .hestia file as string

    Returns
    -------
    pandas.DataFrame
        Pandas dataframe with pivoted array terms
    """
    parsed = json.loads(hestia_file, object_hook=_with_csv_formatting)
    nodes = parsed.get('nodes', [parsed])
    return pivot_nodes(nodes)


def pivot_csv(filepath):
    """
    Pivot terms belonging to array fields of nodes, indexing their value directly with
    the term ID and any distinguishing uniqueness fields in the following format:
    node.arrayField.termId+uniquenessField1[value]+uniquenessField2etc[value]

    Parameters
    ----------
    filepath : str
        Path to the CSV to be formatted.

    Returns
    -------
    pandas.DataFrame
        Pandas dataframe with pivoted array terms
    """
    df_in = pd.read_csv(filepath, index_col=None, dtype=object)
    df_out = _format_and_pivot(df_in)
    return df_out



from numpy import inf, sqrt
from numpy.testing import assert_almost_equal
from pytest import mark

from hestia_earth.utils.stats import (
    truncated_normal_1d,
    _calc_confidence_level, add_normal_distributions, calc_confidence_level_monte_carlo, calc_precision_monte_carlo,
    calc_required_iterations_monte_carlo, calc_z_critical, lerp_normal_distributions, subtract_normal_distributions
)


def test_truncated_normal_1d():
    results = truncated_normal_1d(shape=(1, 10), mu=100, sigma=2, low=0, high=inf, seed=50)[0]
    assert len(results) == 10
    for result in results:
        assert result*0.98 < result < result*1.02, result


# confidence_level, n_sided, z_critical
CONFIDENCE_INTERVAL_PARAMS = [
    # 1 sided
    (0, 1, -inf),
    (50, 1, 0),
    (80, 1, 0.8416),
    (90, 1, 1.2816),
    (95, 1, 1.6449),
    (99, 1, 2.3263),
    (100, 1, inf),
    # 2 sided
    (0, 2, 0),
    (50, 2, 0.6745),
    (80, 2, 1.2816),
    (90, 2, 1.6449),
    (95, 2, 1.9600),
    (99, 2, 2.5758),
    (100, 2, inf)
]


@mark.parametrize(
    "confidence_level, n_sided, z_critical",
    CONFIDENCE_INTERVAL_PARAMS,
    ids=[f"z={z}, n={n}" for _, n, z in CONFIDENCE_INTERVAL_PARAMS]
)
def test_calc_confidence_level(confidence_level, n_sided, z_critical):
    result = _calc_confidence_level(z_critical, n_sided=n_sided)
    assert_almost_equal(result, confidence_level, decimal=2)


@mark.parametrize(
    "confidence_level, n_sided, z_critical",
    CONFIDENCE_INTERVAL_PARAMS,
    ids=[f"conf={conf}, n={n}" for conf, n, _ in CONFIDENCE_INTERVAL_PARAMS]
)
def test_calc_z_critical(confidence_level, n_sided, z_critical):
    result = calc_z_critical(confidence_level, n_sided=n_sided)
    assert_almost_equal(result, z_critical, decimal=4)


# confidence_level, n_iterations, precision, sd
MONTE_CARLO_PARAMS = [
    (95, 80767, 0.01, 1.45),
    (95, 1110, 0.01, 0.17),
    (99, 1917, 0.01, 0.17),
    (50, 102, 100.18, 1500)
]


@mark.parametrize(
    "confidence_level, n_iterations, precision, sd",
    MONTE_CARLO_PARAMS,
    ids=[f"n={n}, prec={prec}, sd={sd}" for _, n, prec, sd in MONTE_CARLO_PARAMS]
)
def test_calc_confidence_level_monte_carlo(confidence_level, n_iterations, precision, sd):
    result = calc_confidence_level_monte_carlo(n_iterations, precision, sd,)
    assert_almost_equal(result, confidence_level, decimal=2)


@mark.parametrize(
    "confidence_level, n_iterations, precision, sd",
    MONTE_CARLO_PARAMS,
    ids=[f"conf={conf}, prec={prec}, sd={sd}" for conf, _, prec, sd in MONTE_CARLO_PARAMS]
)
def test_calc_required_iterations_monte_carlo(confidence_level, n_iterations, precision, sd):
    result = calc_required_iterations_monte_carlo(confidence_level, precision, sd)
    assert result == n_iterations


@mark.parametrize(
    "confidence_level, n_iterations, precision, sd",
    MONTE_CARLO_PARAMS,
    ids=[f"conf={conf}, n={n}, sd={sd}" for conf, n, _, sd in MONTE_CARLO_PARAMS]
)
def test_calc_precision_monte_carlo(confidence_level, n_iterations, precision, sd):
    result = calc_precision_monte_carlo(confidence_level, n_iterations, sd)
    assert_almost_equal(result, precision, decimal=2)


# mu_1, sigma_1, mu_2, sigma_2, rho, sum_mean, sum_sigma, diff_mean, diff_sigma
PARAMS_NORMAL_DIST = [
    # 2 standard normal distributions, perfectly negative correlation
    (0, 1, 0, 1, -1, 0, 0, 0, 2),
    # 2 standard normal distributions, negative correlation
    (0, 1, 0, 1, -0.5, 0, 1, 0, sqrt(3)),
    # 2 standard normal distributions, no correlation
    (0, 1, 0, 1, 0, 0, sqrt(2), 0, sqrt(2)),
    # 2 standard normal distributions, positive correlation
    (0, 1, 0, 1, 0.5, 0, sqrt(3), 0, 1),
    # 2 standard normal distributions, perfectly positive correlation
    (0, 1, 0, 1, 1, 0, 2, 0, 0),
    # different normal distributions, perfectly negative correlation
    (50000, 3000, 45000, 9000, -1, 95000, 6000, 5000, 12000),
    # different normal distributions, no correlation
    (50000, 3000, 45000, 9000, 0, 95000, sqrt(90000000), 5000, sqrt(90000000)),
    # different normal distributions, perfectly positive correlation
    (50000, 3000, 45000, 9000, 1, 95000, 12000, 5000, 6000)
]
IDS_ADD_NORMAL_DIST = [
    f"N({mu_1}, {sigma_1}^2) + N({mu_2}, {sigma_2}^2), rho: {rho}"
    for mu_1, sigma_1, mu_2, sigma_2, rho, *_ in PARAMS_NORMAL_DIST
]
IDS_SUBTRACT_DIST = [
    f"N({mu_1}, {sigma_1}^2) - N({mu_2}, {sigma_2}^2), rho: {rho}"
    for mu_1, sigma_1, mu_2, sigma_2, rho, *_ in PARAMS_NORMAL_DIST
]


@mark.parametrize(
    "mu_1, sigma_1, mu_2, sigma_2, rho, sum_mean, sum_sigma, _diff_mean, _diff_sigma",
    PARAMS_NORMAL_DIST,
    ids=IDS_ADD_NORMAL_DIST
)
def test_add_normal_distributions(mu_1, sigma_1, mu_2, sigma_2, rho, sum_mean, sum_sigma, _diff_mean, _diff_sigma):
    result = add_normal_distributions(mu_1, sigma_1, mu_2, sigma_2, rho)
    assert result == (sum_mean, sum_sigma)


@mark.parametrize(
    "mu_1, sigma_1, mu_2, sigma_2, rho, _sum_mean, _sum_sigma, diff_mean, diff_sigma",
    PARAMS_NORMAL_DIST,
    ids=IDS_SUBTRACT_DIST
)
def test_subtract_normal_distributions(mu_1, sigma_1, mu_2, sigma_2, rho, _sum_mean, _sum_sigma, diff_mean, diff_sigma):
    result = subtract_normal_distributions(mu_1, sigma_1, mu_2, sigma_2, rho)
    assert result == (diff_mean, diff_sigma)


# mu_1, sigma_1, mu_2, sigma_2, alpha, rho, Z_mean, Z_sigma
PARAMS_LERP_NORMAL_DIST = [
    # 2 standard normal distributions, perfectly negative correlation
    (0, 1, 0, 1, 0, -1, 0, 1),
    (0, 1, 0, 1, 0.5, -1, 0, 0),
    (0, 1, 0, 1, 1, -1, 0, 1),
    # 2 standard normal distributions, no correlation
    (0, 1, 0, 1, 0, 0, 0, 1),
    (0, 1, 0, 1, 0.5, 0, 0, sqrt(0.5)),
    (0, 1, 0, 1, 1, 0, 0, 1),
    # 2 standard normal distributions, perfectly positive correlation
    (0, 1, 0, 1, 0, 1, 0, 1),
    (0, 1, 0, 1, 0.5, 1, 0, 1),
    (0, 1, 0, 1, 1, 1, 0, 1),
    # different normal distributions, perfectly negative correlation
    (10000, 3000, 5000, 2500, -0.5, -1, 12500, 5750),
    (10000, 3000, 5000, 2500, 0, -1, 10000, 3000),
    (10000, 3000, 5000, 2500, 0.5, -1, 7500, 250),
    (10000, 3000, 5000, 2500, 1, -1, 5000, 2500),
    (10000, 3000, 5000, 2500, 1.5, -1, 2500, 5250),
    # different normal distributions, no correlation
    (10000, 3000, 5000, 2500, -0.5, 0, 12500, sqrt(21812500)),
    (10000, 3000, 5000, 2500, 0, 0, 10000, 3000),
    (10000, 3000, 5000, 2500, 0.5, 0, 7500, sqrt(3812500)),
    (10000, 3000, 5000, 2500, 1, 0, 5000, 2500),
    (10000, 3000, 5000, 2500, 1.5, 0, 2500, sqrt(16312500)),
    # different normal distributions, perfectly positive correlation
    (10000, 3000, 5000, 2500, -0.5, 1, 12500, 3250),
    (10000, 3000, 5000, 2500, 0, 1, 10000, 3000),
    (10000, 3000, 5000, 2500, 0.5, 1, 7500, 2750.0),
    (10000, 3000, 5000, 2500, 1, 1, 5000, 2500),
    (10000, 3000, 5000, 2500, 1.5, 1, 2500, 2250)
]
IDS_LERP_NORMAL_DIST = [
    f"N({mu_1}, {sigma_1}^2) - N({mu_2}, {sigma_2}^2), alpha: {alpha}, rho: {rho}"
    for mu_1, sigma_1, mu_2, sigma_2, alpha, rho, *_ in PARAMS_LERP_NORMAL_DIST
]


@mark.parametrize(
    "mu_1, sigma_1, mu_2, sigma_2, alpha, rho, Z_mean, Z_sigma",
    PARAMS_LERP_NORMAL_DIST,
    ids=IDS_LERP_NORMAL_DIST
)
def test_lerp_normal_distributions(mu_1, sigma_1, mu_2, sigma_2, alpha, rho, Z_mean, Z_sigma):
    result = lerp_normal_distributions(mu_1, sigma_1, mu_2, sigma_2, alpha, rho)
    assert result == (Z_mean, Z_sigma)



from unittest.mock import patch
import os
import requests
import json
from hestia_earth.schema import SchemaType
import pytest

from .utils import fixtures_path
from hestia_earth.utils.request import api_url
from hestia_earth.utils.api import (
    search,
    find_related,
    download_hestia,
    node_exists,
    find_node,
    find_node_exact,
    find_term_ids_by_names,
)


fake_related_response = {'results': [[{'@id': 'related_id'}]]}
fake_download_response = {'@id': 'id', '@type': 'type'}


class FakeFindRelatedSuccess():
    def json():
        return fake_related_response


class FakeFindRelatedError():
    def json():
        return {}


class FakeFindRelatedException():
    def json():
        raise requests.exceptions.RequestException('error')


class FakeDownloadSuccess():
    def json():
        return fake_download_response


class FakeDownloadError():
    def json():
        raise requests.exceptions.RequestException('error')


class FakeNodeExistSuccess():
    def json():
        return fake_download_response


class FakeNodeExistError():
    def json():
        return {"message": "not-found", "details": {}}


class FakeElasticSearchEmptyResult:
    def json():
        return {"results": []}


@patch('requests.get', return_value=FakeFindRelatedSuccess)
def test_find_related_success(mock_get):
    res = find_related(SchemaType.CYCLE, 'id', SchemaType.SOURCE)
    assert res == fake_related_response.get('results')
    mock_get.assert_called_once_with(
        f"{api_url()}/cycles/id/sources?limit=100", headers={'Content-Type': 'application/json'}
    )


@patch('requests.get', return_value=FakeFindRelatedError)
def test_find_related_error(*args):
    res = find_related(SchemaType.CYCLE, 'id', SchemaType.SOURCE)
    assert not res


@patch('requests.get', return_value=FakeFindRelatedException)
def test_find_related_exception(*args):
    res = find_related(SchemaType.CYCLE, 'id', SchemaType.SOURCE)
    assert not res


@patch('requests.get', return_value=FakeDownloadError)  # make sure fallback is not enabled
def test_download_hestia_local_file(*args):
    id = 'sandContent'
    with open(f"{fixtures_path}/Term/{id}.jsonld", encoding='utf-8') as f:
        expected = json.load(f)
    os.environ['DOWNLOAD_FOLDER'] = fixtures_path
    res = download_hestia(id, SchemaType.TERM)
    assert res == expected
    del os.environ['DOWNLOAD_FOLDER']


@patch('requests.get', return_value=FakeDownloadSuccess)
def test_download_hestia_success(mock_get):
    res = download_hestia('id', SchemaType.SOURCE)
    assert res == fake_download_response
    mock_get.assert_called_once_with(
        f"{api_url()}/sources/id", headers={'Content-Type': 'application/json'})


@patch('requests.get', return_value=FakeDownloadError)
def test_download_hestia_error(*args):
    res = download_hestia('id', SchemaType.SOURCE)
    assert not res


@patch('requests.get', return_value=FakeNodeExistError)  # make sure fallback is not enabled
def test_node_exists_local_file(*args):
    os.environ['DOWNLOAD_FOLDER'] = fixtures_path
    id = 'sandContent'
    assert node_exists(id, SchemaType.TERM)
    del os.environ['DOWNLOAD_FOLDER']


@patch('requests.get', return_value=FakeNodeExistSuccess)
def test_node_exists_true(*args):
    assert node_exists('id', SchemaType.SOURCE)


@patch('requests.get', return_value=FakeNodeExistError)
def test_node_exists_false(*args):
    assert not node_exists('id', SchemaType.SOURCE)


def test_search():
    name = 'Wheat'
    res = search(query={
        'bool': {
            'must': [{'match': {'name': name}}]
        }
    }, limit=2)
    assert res[0].get('name').startswith(name)


def test_find_node():
    name = 'Wheat'
    res = find_node(SchemaType.TERM, {'name': name}, 2)
    assert res[0].get('name').startswith(name)


def test_find_node_exact():
    name = 'Wheat'
    res = find_node_exact(SchemaType.TERM, {'name': name})
    assert not res

    name = 'Wheat, grain'
    res = find_node_exact(SchemaType.TERM, {'name': name})
    assert res.get('name') == name


def test_find_term_ids_by_names():
    names = ["Harris Termite Powder", "Wheat, grain", "Urea (kg N)"]
    res = find_term_ids_by_names(names, 2)
    assert res == {
        "Wheat, grain": "wheatGrain",
        "Harris Termite Powder": "harrisTermitePowder",
        "Urea (kg N)": "ureaKgN",
    }


@patch("requests.post", return_value=FakeElasticSearchEmptyResult)
def test_find_term_ids_by_names_error(mock):
    names = ["id_not_found_name"]
    with pytest.raises(Exception, match=names[0]):
        find_term_ids_by_names(names)



from hestia_earth.schema import SiteSiteType

from hestia_earth.utils.lookup_utils import (
    is_model_siteType_allowed,
    is_model_product_id_allowed,
    is_model_measurement_id_allowed,
    is_siteType_allowed,
    # is_site_measurement_id_allowed,  # TODO: test when lookup is added
    is_input_id_allowed,
    is_input_termType_allowed,
    # is_practice_termType_allowed,  # TODO: test when lookup is added
    # is_practice_id_allowed,  # TODO: test when lookup is added
    is_product_id_allowed,
    is_product_termType_allowed,
    is_node_type_allowed,
    # is_transformation_termType_allowed,  # TODO: test when lookup is added
    # is_transformation_id_allowed,  # TODO: test when lookup is added
)


def test_is_model_siteType_allowed():
    model = 'pooreNemecek2018'
    term_id = 'netPrimaryProduction'

    site = {'siteType': SiteSiteType.POND.value}
    assert is_model_siteType_allowed(model, term_id, {'site': site}) is True

    site = {'siteType': SiteSiteType.CROPLAND.value}
    assert not is_model_siteType_allowed(model, term_id, {'site': site})


def test_is_model_product_id_allowed():
    model = 'schmidt2007'
    term_id = 'ch4ToAirWasteTreatment'

    node = {'products': [{'term': {'@id': 'oilPalmMillEffluentWaste'}}]}
    assert is_model_product_id_allowed(model, term_id, node) is True

    node = {'products': [{'term': {'@id': 'wheatGrain'}}]}
    assert not is_model_product_id_allowed(model, term_id, node)


def test_is_model_measurement_id_allowed():
    model = 'ipcc2019'
    term_id = 'organicCarbonPerHa'

    node = {'measurements': [{'term': {'@id': 'mineralSoils'}}]}
    assert is_model_measurement_id_allowed(model, term_id, node) is True

    node = {'measurements': [{'term': {'@id': 'organicSoils'}}]}
    assert not is_model_measurement_id_allowed(model, term_id, node)


def test_is_siteType_allowed():
    node = {'@type': 'Site', 'siteType': 'cropland'}
    assert not is_siteType_allowed(node, 'pastureGrass')

    node = {'otherSites': [{'@type': 'Site', 'siteType': 'permanent pasture'}]}
    assert is_siteType_allowed(node, 'pastureGrass') is True


def test_is_input_id_allowed():
    term_id = 'noxToAirExcreta'

    node = {'inputs': [{'term': {'@id': 'wheatGrain'}}]}
    assert is_input_id_allowed(node, term_id) is True


def test_is_input_termType_allowed():
    term_id = 'noxToAirExcreta'

    node = {'inputs': [{'term': {'termType': 'crop'}}]}
    assert not is_input_termType_allowed(node, term_id)

    node = {'inputs': [{'term': {'termType': 'excreta'}}]}
    assert is_input_termType_allowed(node, term_id) is True


def test_is_product_id_allowed():
    term_id = 'n2OToAirWasteTreatmentDirect'

    node = {'products': [{'term': {'@id': 'wheatGrain'}}]}
    assert not is_product_id_allowed(node, term_id)

    node = {'products': [{'term': {'@id': 'oilPalmFruit'}}]}
    assert is_product_id_allowed(node, term_id) is True


def test_is_product_termType_allowed():
    term_id = 'ch4ToAirAquacultureSystems'

    node = {'products': [{'term': {'termType': 'crop'}}]}
    assert not is_product_termType_allowed(node, term_id)

    node = {'products': [{'term': {'termType': 'liveAquaticSpecies'}}]}
    assert is_product_termType_allowed(node, term_id) is True


def test_is_node_type_allowed():
    term_id = 'ch4ToAirAquacultureSystems'

    assert not is_node_type_allowed({'type': 'Site'}, term_id)

    assert is_node_type_allowed({'type': 'Cycle'}, term_id) is True



from hestia_earth.schema import SiteSiteType, TermTermType

from hestia_earth.utils.emission import cycle_emissions_in_system_boundary, emissions_in_system_boundary

class_path = 'hestia_earth.utils.emission'


def test_cycle_emissions_in_system_boundary_cropland():
    cycle = {'site': {}}

    cycle['products'] = [
        {
            'term': {
                '@id': 'wheatGrain',
                'termType': TermTermType.CROP.value
            }
        }
    ]
    cycle['site']['siteType'] = SiteSiteType.CROPLAND.value
    term_ids = cycle_emissions_in_system_boundary(cycle)
    assert len(term_ids) > 50

    cycle['products'] = [
        {
            'term': {
                '@id': 'ricePlantFlooded',
                'termType': TermTermType.CROP.value
            }
        }
    ]
    cycle['site']['siteType'] = SiteSiteType.CROPLAND.value
    term_ids = cycle_emissions_in_system_boundary(cycle)
    assert len(term_ids) > 50

    # with inputs restriction, we should have less emissions
    cycle['inputs'] = [
        {
            'term': {'termType': 'crop'}
        }
    ]
    cycle['site']['siteType'] = SiteSiteType.CROPLAND.value
    assert len(cycle_emissions_in_system_boundary(cycle)) < len(term_ids)


def test_cycle_emissions_in_system_boundary_animal_housing():
    cycle = {'site': {}}
    cycle['products'] = [
        {
            'term': {
                '@id': 'meatBeefCattleLiveweight',
                'termType': TermTermType.ANIMALPRODUCT.value
            }
        }
    ]
    cycle['site']['siteType'] = SiteSiteType.ANIMAL_HOUSING.value
    term_ids = cycle_emissions_in_system_boundary(cycle)
    assert len(term_ids) > 20


def test_emissions_in_system_boundary():
    term_ids = emissions_in_system_boundary()
    assert len(term_ids) > 50



import json
from unittest.mock import patch

from hestia_earth.utils.term import download_term

class_path = 'hestia_earth.utils.term'
term = {'@type': 'Term', '@id': 'seed'}
termType = 'seed'


@patch(f"{class_path}._load_from_storage", return_value=json.dumps([term]))
def test_download_term_from_glossary(*args):
    assert download_term(term, termType) == term


@patch(f"{class_path}.download_hestia", return_value=term)
@patch(f"{class_path}._load_term_file", return_value={})
def test_download_term_from_api(*args):
    assert download_term(term, termType) == term



from hestia_earth.utils.date import diff_in_years, is_in_days, is_in_months


def test_diff_in_years():
    assert diff_in_years('1990-01-01', '1999-02-01') == 9.1


def test_is_in_days():
    assert not is_in_days('2000')
    assert not is_in_days('2000-01')
    assert is_in_days('2000-01-01')


def test_is_in_months():
    assert not is_in_months('2000')
    assert is_in_months('2000-01')
    assert not is_in_months('2000-01-01')



import numpy

from .utils import fixtures_path
from hestia_earth.utils.lookup import (
    load_lookup,
    column_name,
    get_table_value,
    find_term_ids_by,
    download_lookup,
    extract_grouped_data,
    extract_grouped_data_closest_date,
    _get_single_table_value,
    lookup_term_ids,
    lookup_columns
)


def test_load_lookup_numpy_array():
    lookup = load_lookup(f"{fixtures_path}/lookup.csv")
    assert isinstance(lookup, numpy.recarray)


def test_column_name():
    assert column_name('Maize (corn)') == 'Maize_corn'
    assert column_name('grassland/pasture/meadow') == 'grasslandpasturemeadow'


def test_get_table_value():
    lookup = load_lookup(f"{fixtures_path}/lookup.csv")

    # single column match
    assert get_table_value(lookup, column_name('Col1'), 'val10', column_name('Col3')) == 'val30'
    # multiple column match
    assert get_table_value(lookup, [
        column_name('Col1'),
        column_name('Col2'),
    ], [
        'val10',
        'val21'
    ], column_name('Col3')) == 'val31'
    # no match
    assert not get_table_value(lookup, column_name('Col10'), 'val10', column_name('Col3'))

    # column does not exist
    assert not get_table_value(lookup, [
        column_name('Col1'),
        column_name('Col2'),
    ], [
        'random',
        'val21'
    ], column_name('random'))

    # table does not exist
    assert not get_table_value(None, column_name('Col10'), 'val10', column_name('Col3'))


def test_get_table_value_empty():
    lookup = load_lookup(f"{fixtures_path}/lookup.csv")
    assert get_table_value(lookup, column_name('Col1'), 'val10', column_name('Col4')) is None
    assert get_table_value(lookup, column_name('Col2'), 'val22', column_name('Col1')) is None


def test_find_term_ids_by():
    lookup = download_lookup('crop.csv')
    assert 'wheatGrain' in find_term_ids_by(lookup, column_name('cropGroupingFAO'), 'Temporary crops')


def test_download_lookup_with_index():
    filename = 'crop.csv'
    lookup = download_lookup(filename, keep_in_memory=False, build_index=True)
    assert isinstance(lookup, dict)


def test_download_lookup_without_index():
    filename = 'crop.csv'
    lookup = download_lookup(filename, keep_in_memory=False, build_index=False)
    assert isinstance(lookup, numpy.recarray)


def test_handle_missing_float_value():
    filename = 'measurement.csv'
    lookup = download_lookup(filename)
    assert get_table_value(lookup, 'termid', 'rainfallPeriod', 'maximum') is None


def test_handle_missing_string_value():
    filename = 'crop.csv'
    lookup = download_lookup(filename)
    assert get_table_value(lookup, 'termid', 'fixedNitrogen', 'combustion_factor_crop_residue') is None


def test_handle_missing_lookup_value():
    filename = 'region-crop-cropGroupingFaostatProduction-price.csv'
    lookup = download_lookup(filename)
    assert get_table_value(lookup, 'termid', 'GADM-CYP', column_name('Sugar crops nes')) is None


def test_extract_grouped_data_no_data():
    assert not extract_grouped_data('', '2000')
    assert not extract_grouped_data('-', '2000')


def test_extract_grouped_data():
    data = 'Average_price_per_tonne:106950.5556;1991:-;1992:-'
    assert extract_grouped_data(data, 'Average_price_per_tonne') == '106950.5556'
    assert extract_grouped_data(data, '2010') is None


def test_extract_grouped_data_lookup():
    filename = 'region-crop-cropGroupingFaostatProduction-price.csv'
    lookup = download_lookup(filename)
    data = get_table_value(lookup, 'termid', 'GADM-NPL', column_name('Chick peas, dry'))
    assert extract_grouped_data(data, '2000') is None
    assert extract_grouped_data(data, '2012') is not None

    filename = 'region-animalProduct-animalProductGroupingFAO-price.csv'
    lookup = download_lookup(filename)
    data = get_table_value(lookup, 'termid', 'GADM-NPL', column_name('Eggs from other birds in shell, fresh, n.e.c.'))
    assert extract_grouped_data(data, '2000') is None
    assert extract_grouped_data(data, '2012') is not None


def test_get_single_table_value_float_values():
    filename = 'ecoClimateZone.csv'
    lookup = download_lookup(filename)
    column = column_name('STEHFEST_BOUWMAN_2006_N2O-N_FACTOR')
    assert _get_single_table_value(lookup, column_name('ecoClimateZone'), 11, column) == -0.3022


def test_extract_grouped_data_closest_date_no_data():
    assert not extract_grouped_data_closest_date('', 2000)
    assert not extract_grouped_data_closest_date('-', 2000)


def test_extract_grouped_data_closest_date():
    data = '2000:-;2001:0.1;2002:0.2;2003:0.3;2004:0.4;2005:0.5'
    assert extract_grouped_data_closest_date(data, 2000) == '0.1'
    assert extract_grouped_data_closest_date(data, 2001) == '0.1'
    assert extract_grouped_data_closest_date(data, 2020) == '0.5'


def test_lookup_term_ids():
    assert 'wheatGrain' in lookup_term_ids(download_lookup('crop.csv'))


def test_lookup_columns():
    assert 'termid' in lookup_columns(download_lookup('crop.csv'))



from hestia_earth.schema import TermTermType

from hestia_earth.utils.model import (
    find_term_match, find_primary_product, filter_list_term_type, convert_value, linked_node
)


def test_find_term_match():
    values = [
        {'term': {'@id': 'my term'}, 'methodModel': {'@id': 'model-1'}},
        {'term': {'@id': 'another id'}},
        {'term': {'@id': 'my term'}, 'methodModel': {'@id': 'model-2'}}
    ]
    assert find_term_match(values, 'my term') == values[0]
    assert find_term_match(values, 'my term', match_params={'methodModel.@id': 'model-2'}) == values[2]


def test_find_primary_product():
    primary = {
        '@type': 'Product',
        'primary': True
    }
    cycle = {
        'products': [
            primary,
            {'@type': 'Product'}
        ]
    }
    assert find_primary_product(cycle) == primary


def test_filter_list_term_type():
    values = [
        {
            'term': {
                'termType': TermTermType.CROP.value
            }
        },
        {
            'term': {
                'termType': TermTermType.FUEL.value
            }
        }
    ]
    assert len(filter_list_term_type(values, TermTermType.CROP)) == 1
    assert len(filter_list_term_type(values, [TermTermType.CROP, TermTermType.FUEL])) == 2


def test_convert_value():
    assert convert_value(1, 'm3', 'kg', density=553) == 553
    assert convert_value(1, 'm3', 'L') == 1000
    assert convert_value(553, 'kg', 'm3', density=553) == 1
    assert convert_value(553, 'kg', 'L', density=553) == 1000
    assert convert_value(1000, 'L', 'm3') == 1
    assert round(convert_value(100, 'MJ', 'kWh')) == 28
    assert convert_value(10, 'kWh', 'MJ') == 36


def test_linked_node():
    node = {
        '@type': 'Term',
        '@id': 'term-id',
        'value': 10,
        'name': 'term'
    }
    assert linked_node(node) == {'@type': 'Term', '@id': 'term-id', 'name': 'term'}

    node['@type'] = 'Cycle'
    assert linked_node(node) == {'@type': 'Cycle', '@id': 'term-id'}



from collections.abc import Iterable
from typing import Optional, Union, Any
from enum import Enum
from functools import reduce
from statistics import mode, mean
from hestia_earth.schema import TermTermType

from .lookup import download_lookup, get_table_value, column_name
from .tools import non_empty_list, non_empty_value, flatten
from .emission import cycle_emissions_in_system_boundary
from .model import filter_list_term_type


def get_lookup_value(blank_node: dict, column: str):
    term = blank_node.get('term', {})
    table_name = f"{term.get('termType')}.csv" if term else None
    value = get_table_value(
        download_lookup(table_name), 'termid', term.get('@id'), column_name(column)
    ) if table_name else None
    return value


def group_by_keys(values: list, group_keys: list = ['term']):
    def node_value(value):
        return (
            value.get('@id') if isinstance(value, dict) else
            list(map(node_value, value)) if isinstance(value, list) else
            value
        )

    def run(group: dict, node: dict):
        group_key = '-'.join(flatten(non_empty_list([node_value(node.get(v)) for v in group_keys])))
        group[group_key] = group.get(group_key, []) + [node]
        return group
    return reduce(run, values, {})


class ArrayTreatment(Enum):
    """
    Enum representing different treatments for arrays of values.
    """
    MEAN = 'mean'
    MODE = 'mode'
    SUM = 'sum'
    FIRST = 'first'
    LAST = 'last'


def _should_run_array_treatment(value):
    return isinstance(value, Iterable) and len(value) > 0


DEFAULT_ARRAY_TREATMENT = ArrayTreatment.MEAN
ARRAY_TREATMENT_TO_REDUCER = {
    ArrayTreatment.MEAN: lambda value: mean(non_empty_list(value)) if _should_run_array_treatment(value) else None,
    ArrayTreatment.MODE: lambda value: mode(non_empty_list(value)) if _should_run_array_treatment(value) else None,
    ArrayTreatment.SUM: lambda value: sum(non_empty_list(value)) if _should_run_array_treatment(value) else None,
    ArrayTreatment.FIRST: lambda value: value[0] if _should_run_array_treatment(value) else None,
    ArrayTreatment.LAST: lambda value: value[-1] if _should_run_array_treatment(value) else None
}
"""
A dictionary mapping ArrayTreatment enums to corresponding reducer functions.
"""


def _retrieve_array_treatment(
    node: dict,
    is_larger_unit: bool = False,
    default: ArrayTreatment = ArrayTreatment.MEAN
) -> ArrayTreatment:
    """
    Retrieves the array treatment for a given node.

    Array treatments are used to reduce an array's list of values into
    a single value. The array treatment is retrieved from a lookup on
    the node's term.

    Parameters
    ----------
    node : dict
        The dictionary representing the node.
    is_larger_unit : bool, optional
        Flag indicating whether to use the larger unit lookup, by default `False`.
    default : ArrayTreatment, optional
        Default value to return if the lookup fails, by default `ArrayTreatment.MEAN`.

    Returns
    -------
    ArrayTreatment
        The retrieved array treatment.

    """
    ARRAY_TREATMENT_LOOKUPS = [
        'arrayTreatmentLargerUnitOfTime',
        'arrayTreatment'
    ]
    lookup = ARRAY_TREATMENT_LOOKUPS[0] if is_larger_unit else ARRAY_TREATMENT_LOOKUPS[1]

    lookup_value = get_lookup_value(node, lookup)

    return next(
        (treatment for treatment in ArrayTreatment if treatment.value == lookup_value),
        default
    )


def get_node_value(
    node: dict,
    key: str = 'value',
    is_larger_unit: bool = False,
    array_treatment: Optional[ArrayTreatment] = None,
    default_array_treatment: Optional[ArrayTreatment] = ArrayTreatment.MEAN,
    default: Any = 0
) -> Union[float, bool]:
    """
    Get the value from the dictionary representing the node,
    applying optional array treatment if the value is a list.

    Parameters
    ----------
    node : dict
        The dictionary representing the node.
    key : str
        The key to retrieve the value for. Will use `value` by default.
    is_larger_unit : bool, optional
        A flag indicating whether the unit of time is larger, by default `False`.
    array_treatment : ArrayTreatment, optional
        Override any array treatment set in the term lookup.
    default_array_treatment : ArrayTreatment, optional
        The default treatment to use when the term has none, and `array_treatment` is not set
    default : any
        The default value, if no value is found or it could not be parsed.

    Returns
    -------
    float | bool
        The extracted value from the node.
    """
    value = node.get(key)

    reducer = ARRAY_TREATMENT_TO_REDUCER[(
        array_treatment or
        _retrieve_array_treatment(node, is_larger_unit=is_larger_unit, default=default_array_treatment)
    )] if isinstance(value, list) and len(value) > 0 else None

    return reducer(value) if reducer else (
        value if any([isinstance(value, float), isinstance(value, int), isinstance(value, bool)]) else
        default if not non_empty_value(value) else
        value
    )


_BLANK_NODE_GROUPING_KEYS = {
    TermTermType.EMISSION: ['methodModel']
}


def get_blank_nodes_calculation_status(node: dict, list_key: str, termType: TermTermType):
    """
    Get calculation status for a Node and a list of Blank node.
    Example: get the calculation status for all emissions included in the HESTIA system boundary.

    Parameters
    ----------
    node : dict
        The dictionary representing the node.
    list_key : str
        The key where the blank nodes are contained. Example: `emissions`.
    termType : TermTermType
        The `term.termType` resitrction for blank nodes. Example for `list_key=emissions`: `TermTermType.EMISSION`.

    Returns
    -------
    dict
        A dictionary of `key:value` pairs representing each `term.@id` found in the blank nodes as key,
        and the resulting calculation as value, containing the recalculated `value`, `method` and `methodTier`.
        Note: if a calculation fails for a blank node, the `value` is an empty dictionary.
    """
    all_term_ids = cycle_emissions_in_system_boundary(node, termType=termType)
    blank_nodes = filter_list_term_type(node.get(list_key, []), termType)
    blank_nodes_by_term = group_by_keys(blank_nodes, ['term'])
    blank_nodes_grouping_keys = _BLANK_NODE_GROUPING_KEYS.get(termType) or []

    def blank_node_data(blank_nodes: list):
        value = get_node_value({
            'term': blank_nodes[0].get('term'),
            'value': list(map(get_node_value, blank_nodes))
        })
        inputs = flatten(map(lambda v: v.get('inputs', []), blank_nodes))
        return {
            'value': value
        } | ({
            'inputs': sorted(list(map(lambda v: v.get('@id'), inputs)))
        } if inputs else {})

    def map_blank_node(term_id: str):
        values = blank_nodes_by_term.get(term_id, [])
        grouped_blank_nodes = group_by_keys(values, blank_nodes_grouping_keys) if blank_nodes_grouping_keys else {}
        return (
            {} if not values else
            {
                k: blank_node_data(v)
                for k, v in grouped_blank_nodes.items()
            } if grouped_blank_nodes else
            blank_node_data([values[0]])
        )

    return {term_id: map_blank_node(term_id) for term_id in all_term_ids}



import os
import json
import pytest
from unittest.mock import Mock, patch, call

from .utils import fixtures_path
from hestia_earth.utils.tools import flatten
from hestia_earth.utils.pipeline import (
    METADATA_PROGRESS_KEY, METADATA_STAGE_KEY,
    parse_event, get_related_nodes, get_related_nodes_data, is_calculating, set_calculating, get_stage, to_string
)

class_path = 'hestia_earth.utils.pipeline'
fixtures_folder = os.path.join(fixtures_path, 'pipeline')


def test_parse_event_direct():
    with open(os.path.join(fixtures_folder, 'direct-invokation.json')) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [{
        'bucket': 'hestia-data-dev',
        'key': 'recalculated/Cycle/1.jsonld'
    }]


def test_parse_event_sqs():
    with open(os.path.join(fixtures_folder, 'sqs.json')) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [{
        'bucket': 'hestia-data-staging',
        'key': 'ImpactAssessment/d3wqcuk_o-vz.jsonld',
        'attributes': {'id': 'd3wqcuk_o-vz', 'type': 'ImpactAssessment'}
    }]


def test_parse_event_sqs_exhausted():
    with open(os.path.join(fixtures_folder, 'sqs-exhausted.json')) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == []


def test_parse_event_sns():
    with open(os.path.join(fixtures_folder, 'sns.json')) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [{
        'bucket': 'hestia-data-dev',
        'key': 'recalculated/Cycle/1.jsonld',
        'attributes': {'Test': 'TestString', 'TestBinary': 'TestBinary'}
    }]


def test_parse_event_s3():
    with open(os.path.join(fixtures_folder, 's3.json')) as f:
        event = json.load(f)
    results = parse_event(event)
    assert results == [{
        'bucket': 'hestia-data-dev',
        'key': 'recalculated/Cycle/1.jsonld'
    }]


_CACHE = {
    'nested': [
        {'@type': 'Site', '@id': 'nested-site-1'},
        {'@type': 'Cycle', '@id': 'nested-cycle-1'},
        {'@type': 'ImpactAssessment', '@id': 'nested-ia-1'},
    ],
    'related': [
        {'@type': 'Site', '@id': 'related-site-1'},
        {'@type': 'Cycle', '@id': 'related-cycle-1'},
        {'@type': 'ImpactAssessment', '@id': 'related-ia-1'},
    ]
}


@pytest.mark.parametrize(
    'test_name,node_type,related_key,related_type,cache,expected_nodes',
    [
        ('nested-with-cache', 'Site', 'nested', 'Cycle', _CACHE, [{'@type': 'Cycle', '@id': 'nested-cycle-1'}]),
        ('related-with-cache', 'Site', 'related', 'Site', _CACHE, [{'@type': 'Site', '@id': 'related-site-1'}]),
        ('related-empty-cache', 'Site', 'related', 'Cycle', {'related': []}, []),
        ('related-no-cache', 'Site', 'related', 'Cycle', None, _CACHE['related'])
    ]
)
@patch(f"{class_path}.find_related", return_value=_CACHE['related'])
def test_get_related_nodes(
    mock_find_related: Mock,
    test_name: str,
    node_type: str,
    related_key: str,
    related_type: str,
    cache: dict,
    expected_nodes: list
):
    nodes = get_related_nodes(
        node={'@type': node_type},
        related_key=related_key,
        related_type=related_type,
        cache=cache
    )
    assert nodes == expected_nodes, test_name


@pytest.mark.parametrize(
    'test_name,node_type,related_key,related_type,cache,expected_nodes',
    [
        ('nested-with-cache', 'Site', 'nested', 'Cycle', _CACHE, [{'@type': 'Cycle', '@id': 'nested-cycle-1'}]),
        ('related-with-cache', 'Site', 'related', 'Site', _CACHE, [{'@type': 'Site', '@id': 'related-site-1'}]),
        ('related-no-cache', 'Site', 'related', 'Cycle', None, _CACHE['related'])
    ]
)
@patch(f"{class_path}.find_related", return_value=_CACHE['related'])
@patch(f"{class_path}._last_modified", return_value='')
@patch(f"{class_path}.is_calculating", return_value=False)
@patch(f"{class_path}.get_stage", return_value=1)
def test_get_related_nodes_data(
    mock_get_stage: Mock,
    mock_is_calculating: Mock,
    mock_last_modified: Mock,
    mock_find_related: Mock,
    test_name: str,
    node_type: str,
    related_key: str,
    related_type: str,
    cache: dict,
    expected_nodes: list
):
    bucket = 'bucket'
    nodes = get_related_nodes_data(
        bucket_name=bucket,
        node={'@type': node_type},
        related_key=related_key,
        related_type=related_type,
        cache=cache
    )

    assert nodes == [
        node | {
            'indexed_at': '',
            'recalculated_at': '',
            'recalculated_stage': 1,
            'is_calculating': False
        }
        for node in expected_nodes
    ], test_name

    assert mock_get_stage.call_args_list == [
        call('bucket', node)
        for node in expected_nodes
    ], test_name

    assert mock_is_calculating.call_args_list == [
        call('bucket', node)
        for node in expected_nodes
    ], test_name

    assert mock_last_modified.call_args_list == flatten([
        [
            call(bucket='bucket', key=f"{node['@type']}/{node['@id']}.jsonld"),
            call(bucket='bucket', key=f"recalculated/{node['@type']}/{node['@id']}.jsonld")
        ]
        for node in expected_nodes
    ]), test_name


@patch(f"{class_path}._read_metadata")
def test_get_stage(mock_read_metadata: Mock):
    bucket = 'bucket'
    node = {'@type': 'Cycle', '@id': 'cycle'}

    mock_read_metadata.return_value = {METADATA_STAGE_KEY: '1'}
    assert get_stage(bucket, node) == 1

    mock_read_metadata.return_value = {}
    assert get_stage(bucket, node) is None


@patch(f"{class_path}._read_metadata")
def test_is_calculating(mock_read_metadata: Mock):
    bucket = 'bucket'
    node = {'@type': 'Cycle', '@id': 'cycle'}

    mock_read_metadata.return_value = {METADATA_PROGRESS_KEY: 'true'}
    assert is_calculating(bucket, node) is True

    mock_read_metadata.return_value = {METADATA_PROGRESS_KEY: 'false'}
    assert not is_calculating(bucket, node)


@patch(f"{class_path}._update_metadata")
def test_set_calculating(mock_update_metadata: Mock):
    bucket = 'bucket'
    node = {'@type': 'Cycle', '@id': 'cycle'}
    set_calculating(bucket, node, True)
    assert mock_update_metadata.call_args_list == [
        call('bucket', 'Cycle/cycle.jsonld', {METADATA_PROGRESS_KEY: 'true'})
    ]


@pytest.mark.parametrize(
    'data,indent,expected',
    [
        ({'@id': 2}, None, '{"@id": 2}'),
        ({'@id': 2}, 2, '{\n  "@id": 2\n}')
    ]
)
def test_to_string(data: dict, indent, expected: str):
    assert to_string(data, indent=indent) == expected, expected



# import os
# import json

# from tests.utils import fixtures_path
# from hestia_earth.utils.cycle import get_cycle_emissions_calculation_status


# def test_get_cycle_emissions_calculation_status():
#     folder = os.path.join(fixtures_path, 'blank_node', 'calculation_status', 'cycle')

#     with open(f"{folder}/node.jsonld", encoding='utf-8') as f:
#         cycle = json.load(f)

#     with open(f"{folder}/emissions-emission-with-missing-inputs.json", encoding='utf-8') as f:
#         expected = json.load(f)

#     result = get_cycle_emissions_calculation_status(cycle)
#     assert result == expected



import os

BUCKET = os.getenv('AWS_BUCKET')
BUCKET_GLOSSARY = os.getenv('AWS_BUCKET_GLOSSARY')
_s3_client = None  # noqa: F824


# improves speed for connecting on subsequent calls
def _get_s3_client():
    global _s3_client
    import boto3
    _s3_client = boto3.session.Session().client('s3') if _s3_client is None else _s3_client
    return _s3_client


def _get_bucket(glossary: bool = False) -> str:
    return BUCKET_GLOSSARY if glossary else BUCKET


def _load_from_bucket(bucket: str, key: str):
    from botocore.exceptions import ClientError
    try:
        return _get_s3_client().get_object(Bucket=bucket, Key=key)['Body'].read()
    except ClientError:
        return None


def _exists_in_bucket(bucket: str, key: str):
    from botocore.exceptions import ClientError
    try:
        _get_s3_client().head_object(Bucket=bucket, Key=key)
        return True
    except ClientError:
        return False


def _read_size(bucket: str, key: str):
    try:
        return _get_s3_client().head_object(Bucket=bucket, Key=key).get('ContentLength')
    except Exception:
        return 0


def _read_metadata(bucket_name: str, key: str):
    try:
        return _get_s3_client().head_object(Bucket=bucket_name, Key=key).get('Metadata', {})
    except Exception:
        return {}


def _update_metadata(bucket: str, key: str, data: dict = {}):
    try:
        metadata = _read_metadata(bucket, key)
        metadata.update(data)
        _get_s3_client().copy_object(
            Bucket=bucket,
            Key=key,
            CopySource={'Bucket': bucket, 'Key': key},
            Metadata=metadata,
            MetadataDirective='REPLACE'
        )
    except Exception:
        pass


def _last_modified(bucket: str, key: str):
    try:
        return _get_s3_client().head_object(Bucket=bucket, Key=key).get('LastModified')
    except Exception:
        return None


def _upload_to_bucket(bucket: str, key: str, body, content_type: str):
    from botocore.exceptions import ClientError
    try:
        return _get_s3_client().put_object(
            Bucket=bucket,
            Key=key,
            Body=body,
            ContentType=content_type
        )
    except ClientError:
        return None


def _list_bucket_objects(bucket: str, folder: str = ''):
    from botocore.exceptions import ClientError
    try:
        paginator = _get_s3_client().get_paginator('list_objects_v2')
        pages = paginator.paginate(Bucket=bucket, Prefix=folder)

        contents = []
        for page in pages:
            contents.extend(page.get('Contents', []))
        return contents
    except ClientError:
        return []


def _delete_objects(bucket: str, objects: list):
    from botocore.exceptions import ClientError
    try:
        # delete in batch of 1000 max allowed
        batch_size = 1000
        for i in range(0, len(objects), batch_size):
            batch_objects = objects[i:i + batch_size]
            _get_s3_client().delete_objects(
                Bucket=bucket,
                Delete={
                    'Objects': batch_objects,
                    'Quiet': True
                }
            )
    except ClientError:
        return None



# import os
# import json

# from tests.utils import fixtures_path
# from hestia_earth.utils.calculation_status import _emissions_with_status, get_nodes_calculations_status_dataframe

# fixtures_folder = os.path.join(fixtures_path, 'calculation_status')


# def test_emissions_with_status():
#     with open(os.path.join(fixtures_folder, 'nodes.json')) as f:
#         nodes = json.load(f)

#     result = _emissions_with_status(nodes[0])
#     assert result == {
#         'emissions-total': 193,
#         'emissions-complete': 56,
#         'emissions-incomplete': 1,
#         'emissions-missing': 136,
#         'emissions': result['emissions']  # ignore
#     }

#     result = _emissions_with_status(nodes[1])
#     assert result == {
#         'emissions-total': 193,
#         'emissions-complete': 0,
#         'emissions-incomplete': 13,
#         'emissions-missing': 180,
#         'emissions': result['emissions']  # ignore
#     }


# def test_get_nodes_calculations_status_dataframe():
#     with open(os.path.join(fixtures_folder, 'nodes.json')) as f:
#         nodes = json.load(f)

#     expected = open(os.path.join(fixtures_folder, 'result.csv'), 'r').read()

#     df = get_nodes_calculations_status_dataframe(nodes, file_format='csv')
#     assert df.to_csv(None, index=None) == expected
